//
// Copyright 2003-2005 (C) Eiichiro ITO, GHC02331@nifty.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eiichiro ITO, 13 August 2003
// mailto: GHC02331@nifty.com

#ifndef Perception_h_DEFINED
#define Perception_h_DEFINED
#include "command.h"
#include "VisionInfo.h"

static const int MAX_COORD = 10;

enum EAiboVisionRangeObject {
    VR_BALL = 0,
    VR_BEACON_ML, VR_BEACON_MR, VR_BEACON_TL, VR_BEACON_TR,
    VR_GOAL_M, VR_GOAL_T, VR_DUMMY5, VR_DUMMY6,
    MAX_VISION_RANGE_OBJECT
};

struct ObjCoord {
    double x, y;
    int distance;
    int mSec;
};

class FoundHistory {
  private:
    ObjCoord ball[MAX_COORD];
    ObjCoord goalB[MAX_COORD];
    ObjCoord goalY[MAX_COORD];
    int curBall;
    int curGoalB;
    int curGoalY;

  public:
    FoundHistory();
    void Add( CdtInfo* cdtInfo, int nowMSec );
    void Print() const;
    void LastObjXYD( EAiboIdentifiedObject eid, double *x, double *y, int *d ) const;
    void LastObjDSec( EAiboIdentifiedObject eid, int *distance, int *mSec ) const;
};

class Perception {
  private:
    CdtInfo recentCdtInfo;	// ŌɌ̏
    CdtInfo currentCdtInfo;	// Ȍ
    int lastCheckedMSec;	// Ōɏ`FbNs
    int nowMSec;		// ݏ`FbNsĂ鎞
    FoundHistory fHistory;	// {[̗

  public:
    Perception();
    void SetCdtInfo( CdtInfo* _cdtInfo );
    void UpdateWhenFound( FoundInfo* recent, FoundInfo* current );
    void UpdateNowMSec( int msec );
    bool VisionCondition( int code, bool bRedTeam ) const;
    bool VisionRangeCondition( EAiboVisionRangeObject id, double angle1, double angle2, double dist1, double dist2, bool bRedTeam ) const;
    bool VisionNearRangeCondition( enum EAiboVisionRangeObject _id, int x1, int x2, int y1, int y2, bool bRedTeam ) const;
    bool MemoryCondition( int code, bool bRedTeam ) const;
    bool EventCondition( int code ) const;
    int GetCdtInfoSeqno() const;
    char *CdtInfoString();
};
#endif // Perception_h_DEFINED
