//
// Copyright 2002 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#ifndef Network_h_DEFINED
#define Network_h_DEFINED

#include <OPENR/OObject.h>
#include <OPENR/OSubject.h>
#include <OPENR/OObserver.h>
#include "TCPConnection.h"
#include "NetworkConfig.h"
#include "def.h"
#include <list>
#include "VisionInfo.h"
#include "command.h"

enum NetworkState {
    NETWORK_IDLE,
    NETWORK_LISTEN,
};

class Network : public OObject {
public:
    Network();
    virtual ~Network() {}

    OSubject*   subject[numOfSubject];
    OObserver*  observer[numOfObserver];     

    virtual OStatus DoInit   (const OSystemEvent& event);
    virtual OStatus DoStart  (const OSystemEvent& event);
    virtual OStatus DoStop   (const OSystemEvent& event);
    virtual OStatus DoDestroy(const OSystemEvent& event);

#ifdef ERS210
    void ListenCont (void* msg);
    void SendCont   (void* msg);
    void ReceiveCont(void* msg);
    void CloseCont  (void* msg);
#endif // ERS210
#ifdef ERS7
    void ListenCont (ANTENVMSG msg);
    void SendCont   (ANTENVMSG msg);
    void ReceiveCont(ANTENVMSG msg);
    void CloseCont  (ANTENVMSG msg);
#endif // ERS7

    void NotifyImage(const ONotifyEvent& event);
    void NotifyControlResult( const ONotifyEvent& event );

private:
    OStatus Listen (int index);
    OStatus Send   (int index);
    OStatus Receive(int index);
    OStatus Close  (int index);
    OStatus InitTCPConnection(int index);

    void SetSendData(int index);
    void SetReceiveData(int index);
    void SendOkWithTagString( int index, const char *tag, const char *string );
    void SendOkWithTagData( int index, const char *tag, const byte *data, int size );
    void ProcessReceivedMessage( int index, const byte* buf, int size );

    // -----
    antStackRef         ipstackRef;
    TCPConnection       connection[NETWORK_CONNECTION_MAX];
    NetworkState        networkState;
    enum EAiboImageType imageType;
    int                 connectionIndex;
    int                 verboseMode;
};

#endif // Network_h_DEFINED
