//
// Copyright 2002 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#include <string.h>
#include <OPENR/OSyslog.h>
#include <OPENR/OPENRAPI.h>
#include <OPENR/core_macro.h>
#include <ant.h>
#include <EndpointTypes.h>
#include <TCPEndpointMsg.h>
#include "Network.h"

OStatus
Network::Listen(int index)
{
    OSYSDEBUG(("Network::Listen()\n"));

    if (connection[index].state != CONNECTION_CLOSED) return oFAIL;

    //
    // Create endpoint
    //
    antEnvCreateEndpointMsg tcpCreateMsg(EndpointType_TCP,
                                         NETWORK_BUFFER_SIZE * 2);
    tcpCreateMsg.Call(ipstackRef, sizeof(tcpCreateMsg));
    if (tcpCreateMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "Network::Listen()",
                  "Can't create endpoint",
                  index, tcpCreateMsg.error));
        return oFAIL;
    }
    connection[index].endpoint = tcpCreateMsg.moduleRef;

    //
    // Listen
    //
    TCPEndpointListenMsg listenMsg(connection[index].endpoint,
                                   IP_ADDR_ANY, NETWORK_PORT);
    listenMsg.continuation = (void*)index;

    listenMsg.Send(ipstackRef, myOID_,
                   Extra_Entry[entryListenCont], sizeof(listenMsg));
    
    connection[index].state = CONNECTION_LISTENING;

    return oSUCCESS;
}

#ifdef ERS210
void
Network::ListenCont(void* msg)
#endif // ERS210
#ifdef ERS7
void
Network::ListenCont(ANTENVMSG msg)
#endif // ERS7
{
    OSYSDEBUG(("Network::ListenCont()\n"));

#ifdef ERS210
    TCPEndpointListenMsg* listenMsg = (TCPEndpointListenMsg*)msg;
#endif // ERS210
#ifdef ERS7
    TCPEndpointListenMsg* listenMsg
        = (TCPEndpointListenMsg*)antEnvMsg::Receive(msg);
#endif // ERS7
    int index = (int)listenMsg->continuation;

    if (listenMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "Network::ListenCont()",
                  "FAILED. listenMsg->error", listenMsg->error));
        Close(index);
        return;
    }

    connection[index].state = CONNECTION_CONNECTED;

    SetReceiveData(index);
    Receive(index);
}

OStatus
Network::Send(int index)
{
    OSYSDEBUG(("Network::Send %d bytes\n", connection[index].sendSize ));
    //OSYSPRINT(("Network::Send %d bytes\n", connection[index].sendSize ));

    if (connection[index].sendSize == 0 ||
        connection[index].state != CONNECTION_CONNECTED) return oFAIL;

    TCPEndpointSendMsg sendMsg(connection[index].endpoint,
                               connection[index].sendData,
                               connection[index].sendSize);
    sendMsg.continuation = (void*)index;

    sendMsg.Send(ipstackRef, myOID_,
                 Extra_Entry[entrySendCont],
                 sizeof(TCPEndpointSendMsg));

    connection[index].state = CONNECTION_SENDING;
    connection[index].sendSize = 0;
    return oSUCCESS;
}

#ifdef ERS210
void
Network::SendCont(void* msg)
#endif // ERS210
#ifdef ERS7
void
Network::SendCont(ANTENVMSG msg)
#endif // ERS7
{
    OSYSDEBUG(("Network::SendCont()\n"));

#ifdef ERS210
    TCPEndpointSendMsg* sendMsg = (TCPEndpointSendMsg*)msg;
#endif // ERS210
#ifdef ERS7
    TCPEndpointSendMsg* sendMsg
	= (TCPEndpointSendMsg*)antEnvMsg::Receive(msg);
#endif // ERS7
    int index = (int)(sendMsg->continuation);

    if (sendMsg->error != TCP_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s %d",
                  "Network::SendCont()",
                  "FAILED. sendMsg->error", sendMsg->error));
        Close(index);
        return;
    }

    //OSYSPRINT(("sent : %-6.6s\n", connection[index].sendData));

    connection[index].state = CONNECTION_CONNECTED;

    SetReceiveData(index);
    Close(index);
}

OStatus
Network::Receive(int index)
{
    OSYSDEBUG(("Network::Receive()\n"));

    if (connection[index].state != CONNECTION_CONNECTED &&
        connection[index].state != CONNECTION_SENDING) return oFAIL;

    TCPEndpointReceiveMsg receiveMsg(connection[index].endpoint,
                                     connection[index].recvData,
                                     1, connection[index].recvSize);
    receiveMsg.continuation = (void*)index;

    receiveMsg.Send(ipstackRef, myOID_,
                    Extra_Entry[entryReceiveCont], sizeof(receiveMsg));

    return oSUCCESS;
}

#ifdef ERS210
void
Network::ReceiveCont(void* msg)
#endif // ERS210
#ifdef ERS7
void
Network::ReceiveCont(ANTENVMSG msg)
#endif // ERS7
{
    OSYSDEBUG(("Network::ReceiveCont()\n"));

#ifdef ERS210
    TCPEndpointReceiveMsg* receiveMsg = (TCPEndpointReceiveMsg*)msg;
#endif // ERS210
#ifdef ERS7
    TCPEndpointReceiveMsg* receiveMsg
        = (TCPEndpointReceiveMsg*)antEnvMsg::Receive(msg);
#endif // ERS7
    int index = (int)(receiveMsg->continuation);

    if (receiveMsg->error != TCP_SUCCESS) {
	OSYSDEBUG(( "Connection closed(%d)\n", receiveMsg->error ));
        Close(index);
        return;
    }

    if ( receiveMsg->sizeMin < 12 ) {
	// ɬפʼХȿãƤʤä
	OSYSPRINT(( "SHORT(%d,%-16.16s)\n",
		    receiveMsg->sizeMin, connection[index].recvData ));
	connection[index].sendSize = 21;
	memcpy( connection[index].sendData,
		"AIBO00000009*NG*SHORT",
		connection[index].sendSize );
	Send( index );
    } else if ( memcmp( connection[index].recvData, "AIBO", 4 ) ) {
	// Magicפʤä
	OSYSPRINT(( "MAGIC(%-16.16s)\n", connection[index].recvData ));
	connection[index].sendSize = 24;
	memcpy( connection[index].sendData,
		"AIBO00000012*NG*BADMAGIC",
		connection[index].sendSize );
	Send( index );
    } else {
	// ʥåƤ
	char sizestr[9];
	memcpy( sizestr, &connection[index].recvData[4], 8 );
	sizestr[8] = '\0';
	int size = atoi(sizestr);

	OSYSPRINT(( "recvData : %-16.16s\n", connection[index].recvData ));
	OSYSPRINT(( "receiving : %d,%-10.10s\n",
		    size, &connection[index].recvData[4] ));

	ProcessReceivedMessage( index,
				&connection[index].recvData[12],
				atoi(sizestr) );
    }
}

OStatus
Network::Close(int index)
{
    OSYSDEBUG(("Network::Close()\n"));

    if (connection[index].state == CONNECTION_CLOSED ||
        connection[index].state == CONNECTION_CLOSING) return oFAIL;

    TCPEndpointCloseMsg closeMsg(connection[index].endpoint);
    closeMsg.continuation = (void*)index;

    closeMsg.Send(ipstackRef, myOID_,
                  Extra_Entry[entryCloseCont], sizeof(closeMsg));

    connection[index].state = CONNECTION_CLOSING;

    return oSUCCESS;
}

#ifdef ERS210
void
Network::CloseCont(void* msg)
#endif // ERS210
#ifdef ERS7
void
Network::CloseCont(ANTENVMSG msg)
#endif // ERS7
{
    OSYSDEBUG(("Network::CloseCont()\n"));

#ifdef ERS210
    TCPEndpointCloseMsg* closeMsg = (TCPEndpointCloseMsg*)msg;
#endif // ERS210
#ifdef ERS7
    TCPEndpointCloseMsg* closeMsg
        = (TCPEndpointCloseMsg*)antEnvMsg::Receive(msg);
#endif // ERS7
    int index = (int)(closeMsg->continuation);

    connection[index].state = CONNECTION_CLOSED;
    Listen(index);
}

OStatus
Network::InitTCPConnection(int index)
{
    OSYSDEBUG(("Network::InitTCPConnection()\n"));

    connection[index].state = CONNECTION_CLOSED;

    // 
    // Allocate send buffer
    //
    antEnvCreateSharedBufferMsg sendBufferMsg(NETWORK_BUFFER_SIZE);

    sendBufferMsg.Call(ipstackRef, sizeof(sendBufferMsg));
    if (sendBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "Network::InitTCPConnection()",
                  "Can't allocate send buffer",
                  index, sendBufferMsg.error));
        return oFAIL;
    }

    connection[index].sendBuffer = sendBufferMsg.buffer;
    connection[index].sendBuffer.Map();
    connection[index].sendData
        = (byte*)(connection[index].sendBuffer.GetAddress());

    //
    // Allocate receive buffer
    //
    antEnvCreateSharedBufferMsg recvBufferMsg(NETWORK_BUFFER_SIZE);

    recvBufferMsg.Call(ipstackRef, sizeof(recvBufferMsg));
    if (recvBufferMsg.error != ANT_SUCCESS) {
        OSYSLOG1((osyslogERROR, "%s : %s[%d] antError %d",
                  "Network::InitTCPConnection()",
                  "Can't allocate receive buffer",
                  index, recvBufferMsg.error));
        return oFAIL;
    }

    connection[index].recvBuffer = recvBufferMsg.buffer;
    connection[index].recvBuffer.Map();
    connection[index].recvData
        = (byte*)(connection[index].recvBuffer.GetAddress());

    return oSUCCESS;
}

void
Network::SetSendData(int index)
{
    memcpy(connection[index].sendData,
           connection[index].recvData, connection[index].sendSize);
}

void
Network::SetReceiveData(int index)
{
    connection[index].recvSize = NETWORK_BUFFER_SIZE;
}

void
Network::SendOkWithTagString( int index, const char *tag, const char *string )
{
    int len;

    len = strlen(string) + 2;
    sprintf( (char*) connection[index].sendData, "AIBO%08d%-2.2s%s",
	     len, tag, string );
    connection[index].sendSize = 12 + len;
    Send( index );
}

void
Network::SendOkWithTagData( int index, const char *tag, const byte *data, int size )
{
    sprintf( (char*) connection[index].sendData,
	     "AIBO%08d%-2.2s", size + 2, tag );
    memcpy( &connection[index].sendData[14], data, size );

    connection[index].sendSize = 14 + size;
    Send( index );
}
